/* Test of robot clearance (distance-to-nearest-wall) computations.
 *
 * Daniel Lowd, 5/9/03
 */

#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string>
#include <fstream.h>
#include <iostream.h>
#include <stdio.h>
#include <vector>
#include "Timing.hh"
#include "KbdInput.hh"

#define PI 3.14159

#define MIN(a, b) (((a) < (b)) ? (a) : (b)) 
#define MAX(a, b) (((a) > (b)) ? (a) : (b)) 
#define DEGREES(a) ((a)*180.0/PI)

/*
 * global parameters (read from "parameters" file)
 */

double COLOR = 10;               // color of particles before they move
double COLOR2 = 16;               // color of particles after they move
double PARTICLE_RADIUS = 50.0;   // size of drawn particles
double NOISE = 10.0;             // mean of the sensor noise
double STDEV = 10.0;             // st. deviation of the sensor noise
double NPART = 100;              // number of particles used
double POS_STDEV_FRAC = 0.2;     // fraction of robot moves used as
                                 //   the st. dev. of the distribution
                                 //   on where the robot actually goes
/*
 * symbolic constants
 */

const int POSITIVE_X = 0;
const int POSITIVE_Y = 4;
const int NEGATIVE_X = 8;
const int NEGATIVE_Y = 12;


/*
 * the Wall class
 */

class Wall
{
  public:

    double x1,y1,x2,y2;

    Wall(double x1_, double y1_, double x2_, double y2_):
      x1(x1_), y1(y1_), x2(x2_), y2(y2_) { ; }

    Wall():
      x1(.0), y1(.0), x2(.0), y2(.0) { ; }

};

/*
 * the Particle class
 */

class Particle
{
  public:
    double x,y,a,w; // x, y, angle, and weight

    Particle(double x_, double y_, double a_, double w_):
      x(x_), y(y_), a(a_), w(w_) { ; }

    Particle():
      x(.0), y(.0), a(.0), w(.0) { ; }
};


double
clearance(vector<Wall>& w, double x, double y, double a, int dir)
{
    double result = 1.0e9;
    double diff = 0.0;

    // Convert robot angle and sensor direction to an absolute angle
    // in degrees.
    a += dir * 225.0;
    a /= 10.0;

    // Convert from degrees to radians, in the range -PI to PI
    a *= PI/180.0;
    while (a > PI)
        a -= 2*PI;

    // DEBUG
    //printf("a = %f\n", a);

    for (int i=0 ; i<w.size() ; i++)
    {
        // Calculate angles to each wall endpoint, normalized so
        // that 0 degrees is the robot's angle.
        double theta1 = atan2(w[i].y1 - y, w[i].x1 - x) - a;
        double theta2 = atan2(w[i].y2 - y, w[i].x2 - x) - a;

        // DEBUG
        printf("(%f, %f) - (%f, %f)\n", w[i].x1, w[i].y1, w[i].x2, w[i].y2);
        printf("theta1: %f; theta2: %f\n", DEGREES(theta1), DEGREES(theta2));

        // Shift angles by 2PI until they're in the range PI to -PI
        while (theta1 > PI) theta1 -= 2*PI;
        while (theta2 > PI) theta2 -= 2*PI;
        while (theta1 < -PI) theta1 += 2*PI;
        while (theta2 < -PI) theta2 += 2*PI;

        // If the endpoints are both on one side of the robot
        // or behind it, then skip this wall.
        if ((theta1 > 0 && theta2 > 0) ||
            (theta1 < 0 && theta2 < 0) ||
            (fabs(theta1 - theta2) > PI))
        {
            continue;
        }

        // Calculate angle between two wall endpoints
        double other = atan2(w[i].y2 - w[i].y1, w[i].x2 - w[i].x1) - a;
        while (other < 0)  other += PI;
        while (other > PI) other -= PI;

        double angle3 = MAX(theta1, theta2);
        double angle2 = PI - other;
        double angle1 = PI - angle3 - angle2;

        double xdiff2, ydiff2;
        if (theta1 > theta2)
        {
            xdiff2 = w[i].x1 - x;
            ydiff2 = w[i].y1 - y;
        }
        else
        {
            xdiff2 = w[i].x2 - x;
            ydiff2 = w[i].y2 - y;
        }

        // DEBUG
        //printf("%d: %f\t%f\t%f\n", i, DEGREES(angle1), DEGREES(angle2), 
        //        DEGREES(angle3));

        /*
        if (angle1 < 0)
        {
            printf("angle1 = %f\n", angle1);
            continue;
        }

        if (angle2 < 0)
            printf("angle2 negative!\n");
            */

        // Law of sines
        double dist2 = sqrt(xdiff2*xdiff2 + ydiff2*ydiff2);
        double diff = dist2*sin(angle1)/sin(angle2);

        // DEBUG
        // printf("dist2: %f\n", dist2);

        if (diff < result) result = diff;
    }

  return result;
}

void main()
{
    vector<Wall> w(1);

    for (int i = 0; i < 1000; i++)
    {
        int x1 = rand() % 2000 - 1000;
        int y1 = rand() % 2000 - 1000;
        int x2 = rand() % 2000 - 1000;
        int y2 = rand() % 2000 - 1000;

        w[0] = Wall(x1, y1, x2, y2);
        double dist = clearance(w, 0, 0, 0, 0);

        int ymin = MIN(y1, y2);
        int ymax = MAX(y1, y2);
        int xmin = MIN(x1, x2);
        int xmax = MAX(x1, x2);
        if (ymin > 0 || ymax < 0)
        {
            if (dist < 1000000)
            {
                printf("ERROR (%d): y1 = %d; y2 = %d; dist = %f\n\r", i, 
                        y1, y2, dist);
            }
            continue;
        }

        if (xmax < 0)
        {
            if (dist < 1000000)
                printf("ERROR (%d): dist = %f (shouldn't hit: xmax low)\n\r", i, dist);

            continue;
        }

        int xpos = x1 + (x2 - x1)*(0 - y1)/(y2-y1);
        if (xpos < 0)
        {
            if (dist < 1000000)
            {
                printf("ERROR (%d): dist = %f (shouldn't hit: xpos low)\n\r", i, dist);
            }
            continue;
        }

        if (abs(xpos - dist) > 3)
        {
            printf("ERROR (%d): xpos = %d; dist = %f\n\r", i, xpos, dist);
            //printf("ymin: %d; ymax: %d; xmin: %d; xmax: %d\n\r", ymin, ymax,
            //        xmin, xmax);
        }
    }
}
